from typing import Type, TypeVar
from uuid import UUID

from fastapi import HTTPException
from pydantic import BaseModel
from sqlalchemy.orm import Session

from .models.base import Base

T = TypeVar("T", bound=Base)


def get_object_or_404(db_class: Type[T], id: UUID | str, db: Session, expunge: bool = False) -> T:
    obj = db.query(db_class).filter(db_class.id == id).one_or_none()
    if obj is None:
        raise HTTPException(status_code=404, detail="The object does not exist.")
    if expunge:
        db.expunge(obj)
    return obj


# TODO: Add testing
def create_obj_from_data(
    data: BaseModel, model: Type[T], db: Session, additonal_data={}, exclude={}
) -> T:
    obj = model(**data.model_dump(exclude=exclude) | additonal_data)
    db.add(obj)
    db.commit()
    db.refresh(obj)
    return obj


# TODO: Add testing
def update_obj_from_data(
    data: BaseModel,
    model: Type[T],
    id: UUID | str,
    db: Session,
    partial: bool = False,
    ignore_fields=[],
    additional_data={},
    exclude={},
) -> T:
    obj = get_object_or_404(model, id, db)
    data_dict = data.model_dump(exclude_unset=not partial, exclude=exclude)
    data_dict.update(additional_data)  # merge additional_data into data_dict
    for field in data_dict:
        if field not in ignore_fields:
            setattr(obj, field, data_dict[field])
    db.commit()
    db.refresh(obj)
    return obj


# TODO: Add testing
def delete_object(db_class: Type[T], id: UUID | str, db: Session) -> None:
    obj = db.query(db_class).filter(db_class.id == id).one_or_none()
    if obj is None:
        raise HTTPException(status_code=404, detail="The object does not exist.")
    db.delete(obj)
    db.commit()



from math import ceil
from typing import Any, Generic, Optional, Self, Sequence, TypeVar

from fastapi_pagination import Params
from fastapi_pagination.bases import AbstractPage, AbstractParams
from fastapi_pagination.types import GreaterEqualOne, GreaterEqualZero
from pydantic.json_schema import SkipJsonSchema

T = TypeVar("T")


# TODO: Add complete fastapi-pagination proxy here
# TODO: Add pagination off functionality
# SkipJsonSchema is used to avoid generating invalid JSON schema in FastAPI
class Page(AbstractPage[T], Generic[T]):
    results: Sequence[T]
    page: GreaterEqualOne | SkipJsonSchema[None] = None
    size: GreaterEqualOne | SkipJsonSchema[None] = None
    pages: GreaterEqualZero | SkipJsonSchema[None] = None
    total: GreaterEqualZero
    has_next: bool | SkipJsonSchema[None] = None
    has_prev: bool | SkipJsonSchema[None] = None

    __params_type__ = Params

    @classmethod
    def create(
        cls,
        items: Sequence[T],
        params: AbstractParams,
        *,
        total: Optional[int] = None,
        **kwargs: Any,
    ) -> Self:
        if not isinstance(params, Params):
            raise TypeError("Page should be used with Params")

        size = params.size or total or len(items)
        page = params.page or 1
        pages = None

        if total is not None:
            if total == 0:
                pages = 1
            else:
                pages = ceil(total / size)

        has_next = page < (pages or 1)
        has_prev = page > 1

        return cls(
            total=total,
            results=items,
            page=page,
            size=size,
            pages=pages,
            has_next=has_next,
            has_prev=has_prev,
        )


# Parse response from an SDK to a PAGE
def parse_page(response, page: int, size: int) -> Page:
    return Page(
        page=page,
        size=size,
        total=response.total,
        results=response.results,
        pages=response.pages or 1,
        has_next=response.has_next,
        has_prev=response.has_prev,
    )



import uuid
from datetime import datetime

from sqlalchemy import Column, DateTime, String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import as_declarative
from sqlalchemy.sql import func


@as_declarative()
class Base:
    __abstract__ = True
    # Primary key as uuid
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)
    created_by_id = Column(String)

    __name__: str

    # TODO: Add default representation string
    # TODO: Add automated foreign key resolution

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel, ConfigDict


# The created_by_id is a string because we use the sub from Auth0
class BaseSchemaModelIN(BaseModel):
    created_by_id: str
    model_config = ConfigDict(from_attributes=True)


class BaseSchemaModelOUT(BaseSchemaModelIN):
    id: UUID
    created_at: datetime
    updated_at: datetime



import json

from fastapi.testclient import TestClient


class GenericClient(TestClient):
    def __init__(self, client):
        self.c = TestClient(client)

    def get(self, url: str, r_code: int = 200, parse_json=True):
        re = self.c.get(url)
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if parse_json else re.content

    def delete(self, url: str, r_code: int = 204):
        re = self.c.delete(url)
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if r_code != 204 else None

    def post(
        self, url: str, obj: dict = {}, r_code: int = 201, raw_response=False, *args, **kwargs
    ):
        re = self.c.post(
            url, data=json.dumps(obj), headers={"Content-Type": "application/json"}, *args, **kwargs
        )
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    def post_file(self, url: str, file, r_code: int = 201, raw_response=False, *args, **kwargs):
        re = self.c.post(url, files={"file": file}, *args, **kwargs)
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    def patch(
        self, url: str, obj: dict = {}, r_code: int = 200, raw_response=False, *args, **kwargs
    ):
        re = self.c.patch(
            url, data=json.dumps(obj), headers={"Content-Type": "application/json"}, *args, **kwargs
        )
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    def put(self, url: str, obj: dict = {}, r_code: int = 200, raw_response=False, *args, **kwargs):
        re = self.c.put(
            url,
            data=json.dumps(obj),
            headers={
                "Content-Type": "application/json",
                "accept": "application/json",
            },
            *args,
            **kwargs,
        )
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    def obj_lifecycle(self, input_obj: dict, url: str, pagination: bool = True):
        # GET LIST
        re = self.get(url)
        if pagination:
            assert re["total"] == 0
            assert len(re["results"]) == 0
        else:
            assert len(re) == 0

        # CREATE
        re = self.post(url, obj=input_obj)
        assert "id" in re
        assert re["id"] is not None
        assert re["created_at"] is not None

        obj_id = str(re["id"])

        # GET
        re = self.get(f"{url}{obj_id}/")
        assert re["id"] == obj_id

        # GET LIST
        re = self.get(url)
        if pagination:
            assert re["total"] == 1
            assert len(re["results"]) == 1
        else:
            assert len(re) == 1

        # DELETE
        self.delete(f"{url}{obj_id}")

        # GET LIST
        re = self.get(url)
        if pagination:
            assert re["total"] == 0
            assert len(re["results"]) == 0
        else:
            assert len(re) == 0

        # GET
        self.get(f"{url}{obj_id}", parse_json=False, r_code=404)



import os
from typing import Generator

from dotenv import load_dotenv
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

load_dotenv()

host = os.getenv("POSTGRES_HOST", "localhost")
user = os.getenv("POSTGRES_USER", "postgres")
password = os.getenv("POSTGRES_PASSWORD", "root")
port = os.getenv("POSTGRES_PORT", "5432")
name = os.getenv("POSTGRES_DB", "fastapi")

SQLALCHEMY_DATABASE_URL = f"postgresql+psycopg://{user}:{password}@{host}:{port}/"


engine = create_engine(SQLALCHEMY_DATABASE_URL + name, pool_pre_ping=True)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def get_db() -> Generator[Session, None, None]:
    with SessionLocal() as db:
        yield db



import subprocess

from setuptools import find_packages, setup

with open("requirements.txt") as f:
    requirements = f.read().splitlines()


def get_latest_git_tag() -> str:
    try:
        tag = (
            subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"])
            .strip()
            .decode("utf-8")
        )
        if tag.startswith("v"):
            tag = tag[1:]
    except subprocess.CalledProcessError:
        raise RuntimeError("Unable to get latest git tag")
    return str(tag)


setup(
    name="creyPY",
    version=get_latest_git_tag(),
    description="Collection of my Python and FastAPI shortcuts, snippets etc.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Conrad Großer",
    author_email="conrad@noah.tech",
    packages=find_packages(),
    url="https://github.com/creyD/creyPY",
    license="MIT",
    python_requires=">=3.12",
    install_requires=requirements,
    keywords=[
        "creyPY",
        "Python",
        "FastAPI",
        "shortcuts",
        "snippets",
        "utils",
        "personal library",
    ],
    platforms="any",
)
